///////////////////////////////////////
//  copyright Alexandre Chouvellon 2009
//  Source libre de droit
///////////////////////////////////////

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "analysis.h"
#include "tools.h"
#include "list.fwd.h"
#include "location.h"
#include "board.h"
#include "language.fwd.h"
#include "timer.h"

void getAppropriateList(Board *board,  ChainedList *** a);
int analyzeAroundBlank(Element * element, Action forbiddenAction, int id);
extern Token glb_tokens[12];

//historique des configurations dj tudie
ChainedList   *glb_list00 = NULL;   //regroupe les config avec la pice carre en 0 0 
ChainedList   *glb_list01 = NULL;   //   .....                                    0 1
ChainedList   *glb_list02 = NULL;
ChainedList   *glb_list10 = NULL;
ChainedList   *glb_list11 = NULL;
ChainedList   *glb_list12 = NULL;
ChainedList   *glb_list20 = NULL;
ChainedList   *glb_list21 = NULL;
ChainedList   *glb_list22 = NULL;
ChainedList   *glb_list30 = NULL;
ChainedList   *glb_list31 = NULL;
ChainedList   *glb_list32 = NULL;
MyFifoList    *glb_myFifo = NULL;

void freeAllChainedLists()
{
	freeChainedList(glb_list00);
	freeChainedList(glb_list01);
	freeChainedList(glb_list02);
	freeChainedList(glb_list10);
	freeChainedList(glb_list11);
	freeChainedList(glb_list12);
	freeChainedList(glb_list20);
	freeChainedList(glb_list21);
	freeChainedList(glb_list22);
	freeChainedList(glb_list30);
	freeChainedList(glb_list31);
	freeChainedList(glb_list32);
}

//tst si le pice carr est devant la sortie
int isWinner(Board *board)
{
	return getIdfromBoard(board, 4,2) == DONKEY && getIdfromBoard(board, 4,1) == DONKEY && getIdfromBoard(board, 3,1) == DONKEY && getIdfromBoard(board, 3,2) == DONKEY; 
}

//fonction de construction d'une nouvelle configuration
//push dans l'historique et la fifo si ncessaire
int updateAfterMove(Board * board, Action action, Location * blank1, Location * blank2, Element * previousState)
{
	Element * element2 = NULL;
	ChainedList ** a  =NULL;
	int solutionFound = FALSE;
	char added = FALSE;

	getAppropriateList(board, &a);
	*a = pushMe(*a, board, action, &element2, &added);

	if (isWinner(getBoard(element2)))
	{
		float b= stopTimer(); 
		int deep =0;

		solutionFound = TRUE;
		setPreviousElement(element2, previousState);
		setActionDone(element2, action);
		//la solution est affiche
		deep = explainSolution(element2, deep);
		printf(glb_found, b, deep);
	}else
	{
		if (added)//est ce une nouvelle configuration?
		{
			setPreviousElement(element2, previousState);
			copyLocation(blank1,getBlank1(element2));
			copyLocation(blank2,getBlank2(element2));
			setActionDone(element2, action);
			pushToEnd(&glb_myFifo, element2);
		}
	}

	return solutionFound;
}

//fonction permettant de creer une nouvelle configuration  partir d'un tat initiale et d'une action
//element : configuration initiale
//id: l'id de la pice que l'on va bouger
//orientation : la direction de l'action
//action:  info redondante sur l'id et orientation mais evite un calcul superflu
//realLocation: l'emplacement de la pice que l'on va bouger, pour les pices non unitaires, la partie la plus en haut  gauche est pris en compte 
int analyseSingle(Element * element, short int id, short orientation, Action action, Location * realLocation)
{
	int solutionFound = FALSE;
	int res = FALSE;
	Location blank1 = {0};
	Location blank2 = {0};
	Board board = {0};

	//on copie l'tat courant pour viter de l'endommager dans notre recherche
	copyBoard(getBoard(element), &board);
	copyLocation(getBlank1(element), &blank1);
	copyLocation(getBlank2(element), &blank2);

	//on prpare la board  un dplacement en enlevant l'id du vide 2 de cette derniere
	//en effet, le code de dplacement s'atten  trouver 0 dasn les zones vides
	setOrRemoveId(&board,BLANK2, blank2.row , blank2.column, 0);

	res = moveToken(id, orientation, &board, realLocation->row, realLocation->column, &blank1, &blank2);

	//on resete l'id du blank2 pour revenir  un tableau cohrent avec les spcifications
	setOrRemoveId(&board,BLANK2, blank2.row , blank2.column, 1);

	if (res) //ca a boug, il faut pusher notre lment si il n'a pas encore t tudi
	{  
		solutionFound = updateAfterMove(&board, action, &blank1, &blank2, element);
	}

	return solutionFound;
}

int analyzeAroundBlanks(Element * element)
{
	//l'action inverse de celle qui a permis d'arriver  l'tat actuel doit tre ignor
	Action forbidden = inverseAction(getActionDone(element));
	return analyzeAroundBlank(element, forbidden, BLANK1) || analyzeAroundBlank(element, forbidden, BLANK2);
}

//recupere l'emplacement de l'lment de la pice le plus en haut  gauche
void seekRealLocation(Board * board, char row , char col , Location * real)
{
	short id = getIdfromBoard(board,row, col);
	int type = getTokenTypeFromId(id);

	real->row = row;
	real->column = col;

	if (type == SIMPLE)
	{
		return;
	}

	if (type == DOUBLE_HORIZONTAL)
	{
		if (getIdfromBoard(board,row, col-1) == id)
		{
			real->column = col-1;
		}
		return;
	}

	if (type == DOUBLE_VERTICAL)
	{
		if (getIdfromBoard(board,row-1, col) == id)
		{
			real->row = row-1;
		}
		return;
	}

	if (type == QUAD)
	{
		if (getIdfromBoard(board,row-1, col)!= id &&  getIdfromBoard(board,row, col-1)!= id) //haut gauche
		{
			return;
		}

		if (getIdfromBoard(board,row-1, col)!= id &&  getIdfromBoard(board,row, col-1) == id)//haut droite
		{
			real->column = col-1;
			return;
		}

		if (getIdfromBoard(board,row-1, col)== id &&  getIdfromBoard(board,row, col-1)== id)//bas droite
		{
			real->column = col-1;
			real->row = row-1;
			return;
		}
		if (getIdfromBoard(board,row-1, col)== id &&  getIdfromBoard(board,row, col-1)!= id) //bas gauche
		{
			real->row = row-1;
			return;
		}
		return;
	}
	return ;
}

//tude des mouvement des pices autour de vide ayant l'id idBlank
//element: la configuration actuelle des pices
//forbidddenAction : l'action inverse ayant permise cette configuration
//idBlank:  l'id du vide actuellement tudi
//return TRUE si une configuration tudie est gagnante
int analyzeAroundBlank(Element * element, Action forbiddenAction, int idBlank)
{  
	short idToken = 0;
	int solutionFound = FALSE;
	Action action;
	Board * board = getBoard(element);
	Location realLocation = {0};
	Location * location =(idBlank==BLANK1)?getBlank1(element):getBlank2(element);
	char skipped = FALSE;

	//la piece de gauche
	if (location->column -1 >= 0 )  
	{   
		idToken = getIdfromBoard(board, location->row ,location->column-1);
		if( idToken != BLANK1 && idToken != BLANK2)
		{
			action = RIGHT << 4 | idToken;
			if (idBlank == BLANK2)
			{
				switch (getTokenTypeFromId(idToken))
				{
				case DOUBLE_VERTICAL:			skipped = TRUE; break;
				case QUAD:						skipped = TRUE;	break;
				}	 }
			if (action != forbiddenAction && !skipped)
			{
				seekRealLocation(board, location->row ,location->column-1, &realLocation);
				solutionFound = analyseSingle(element, idToken, RIGHT, action, &realLocation);
			}
		}
	}
	skipped = FALSE;
	// la piece de droite existe?
	if (location->column +1 < 4 && !solutionFound)   
	{
		idToken = getIdfromBoard(board, location->row,location->column +1);
		if( idToken != BLANK1 && idToken != BLANK2)
		{
			action = LEFT << 4 | idToken;
			if (idBlank == BLANK2)
			{
				switch (getTokenTypeFromId(idToken))
				{
				case DOUBLE_VERTICAL:			skipped = TRUE; break;
				case QUAD:						skipped = TRUE;	break;
				}
			}
			if (action != forbiddenAction && !skipped)
			{
				seekRealLocation(board,location->row ,location->column+1, &realLocation);
				solutionFound = analyseSingle(element, idToken, LEFT, action, &realLocation);
			}
		}
	}
	skipped = FALSE;  
	//la piece du haut
	if (location->row -1 >= 0&& !solutionFound)   
	{
		idToken = getIdfromBoard(board, location->row -1 , location->column);
		if( idToken != BLANK1 && idToken != BLANK2)
		{
			action = BOTTOM << 4 | idToken;
			if (idBlank == BLANK2)
			{	
				switch (getTokenTypeFromId(idToken))
				{
				case DOUBLE_HORIZONTAL:			skipped = TRUE; break;
				case QUAD:						skipped = TRUE;	break;
				}
			}
			if (action != forbiddenAction && !skipped)
			{
				seekRealLocation(board, location->row-1 ,location->column, &realLocation);
				solutionFound = analyseSingle(element, idToken, BOTTOM, action, &realLocation);
			}
		}
	}
	skipped = FALSE;
	//la piece du bas
	if (location->row +1 < 5 && !solutionFound)  
	{    
		idToken = getIdfromBoard(board, location->row +1 , location->column);

		if( idToken != BLANK1 && idToken != BLANK2)
		{
			action = TOP << 4 | idToken;

			if (idBlank == BLANK2)
			{
				switch (getTokenTypeFromId(idToken))
				{
				case DOUBLE_HORIZONTAL:			skipped = TRUE; break;
				case QUAD:						skipped = TRUE;	break;
				}
			}
			if (action != forbiddenAction && !skipped)
			{
				seekRealLocation(board,location->row+1 ,location->column, &realLocation);
				solutionFound = analyseSingle(element, idToken, TOP, action, &realLocation);
			}
		}
	}

	return solutionFound;
}

//donne la liste chane trie approprie par rapport  la position de la pice carr, on pourrait continuer le dcoupage 
void getAppropriateList(Board *board,  ChainedList *** a)
{
	char row =0;
	char column =0;
	int found= FALSE; 

	for (row=0; row < 4 ; ++row)
	{
		for (column=0; column < 3 ; ++column)
		{
			found = getIdfromBoard(board, row,column) == DONKEY;
			if (found) break;
		}
		if (found) break;
	}
	switch(row)
	{
	case 0:
		switch(column)
		{
		case 0: *a= &glb_list00;  break;
		case 1: *a= &glb_list01;  break;
		case 2: *a= &glb_list02;  break;
		}break;
	case 1:
		switch(column)
		{
		case 0: *a= &glb_list10;  break;
		case 1: *a= &glb_list11;  break;
		case 2: *a= &glb_list12;  break;
		}break;
	case 2: 
		switch(column)
		{
		case 0: *a= &glb_list20;  break;
		case 1: *a= &glb_list21;  break;
		case 2: *a= &glb_list22;  break;
		}break;
	case 3: 
		switch(column)
		{
		case 0: *a= &glb_list30;  break;
		case 1: *a= &glb_list31;  break;
		case 2: *a= &glb_list32;  break;
		}break;
	}
}


//boucle de recherche principale
//tant que la fifo n'est pas vide  et qu'aucune solution n'a encore t trouv, on continue la recherche
//le programme termine forcment puisque le nombre de combinaison  tudier est fini.
int analyseStateFIFO()
{
	ChainedList * current = NULL;
	int solutionFound = FALSE;

	while( (current = pop(glb_myFifo)) != NULL && !solutionFound)
	{
		//on prend le premier lment de la liste, on tudie tous les mouvements possibles, on les push dans l'historique et si l'lment est nouveau on le rajoute dans la fifo
		if(current)
		{
			solutionFound = analyzeAroundBlanks(getElement(current));
			free(current);
		}
	}
	return solutionFound;
}
